-----------------------------------------------------------------------------------------------
-- QtPerformance by Ninix
-----------------------------------------------------------------------------------------------
 
require "Window"
 

local QtPerformance = {} 

local VERSION_NOTE = "v1.1 [12 October 2015]"
 
local VIDEO_PRESETS_DX11 = {
	["Medium"] =
	{
		["video.framerateMax"] = 100,
		["lod.viewDistance"] = 640,
		["lod.farFogDistance"] = 2048,
		["camera.distanceMax"] = 32,
		["lod.textureLodMin"] = 1,
		["gfx.jpgdstfmt"] = 0,
		["lod.textureFilter"] = 2,
		["lod.landLod"] = 1,
		["lod.clutterDistance"] = 64,
		["lod.clutterDensity"] = 1,
		["draw.shadows"] = true,
		["lod.shadowMapSize"] = 2048,
		["lod.renderTargetScale"] = 1,
		["fxaa.preset"] = 3,
		["fxaa.enable"] = true,
		["world.propScreenHeightPercentMin"] = 8,
		["particle.envParticleScale"] = 0.5,
		["spell.visualSuppression"] = 0.5,
		["spell.visualSuppressedAlpha"] = 0,
		["ui.enableAddonProfiling"] = false,
	},
	["Low"] =
	{
		["video.framerateMax"] = 100,
		["lod.viewDistance"] = 512,
		["lod.farFogDistance"] = 1536,
		["camera.distanceMax"] = 32,
		["lod.textureLodMin"] = 2,
		["gfx.jpgdstfmt"] = 0,
		["lod.textureFilter"] = 1,
		["lod.landLod"] = 0,
		["lod.clutterDistance"] = 48,
		["lod.clutterDensity"] = 0,
		["draw.shadows"] = true,
		["lod.shadowMapSize"] = 1024,
		["lod.renderTargetScale"] = 0.75,
		["fxaa.preset"] = 1,
		["fxaa.enable"] = true,
		["world.propScreenHeightPercentMin"] = 12,
		["particle.envParticleScale"] = 0.25,
		["spell.visualSuppression"] = 0.9,
		["spell.visualSuppressedAlpha"] = 0,
		["ui.enableAddonProfiling"] = false,
	},
	["UltraLow"] =
	{
		["video.framerateMax"] = 100,
		["lod.viewDistance"] = 256,
		["lod.farFogDistance"] = 1024,
		["camera.distanceMax"] = 32,
		["lod.textureLodMin"] = 2,
		["gfx.jpgdstfmt"] = 0,
		["lod.textureFilter"] = 0,
		["lod.landLod"] = 0,
		["lod.clutterDistance"] = 32,
		["lod.clutterDensity"] = 0,
		["draw.shadows"] = false,
		["lod.shadowMapSize"] = 1024,
		["lod.renderTargetScale"] = 0.5,
		["fxaa.preset"] = 0,
		["fxaa.enable"] = false,
		["world.propScreenHeightPercentMin"] = 12,
		["particle.envParticleScale"] = 0.1,
		["spell.visualSuppression"] = 1.0,
		["spell.visualSuppressedAlpha"] = 0,
		["ui.enableAddonProfiling"] = false,
	}
}

local VIDEO_PRESETS_DX9 ={
	["Medium"] =
	{
		["video.framerateMax"] = 100,
		["lod.viewDistance"] = 384,
		["lod.farFogDistance"] = 2048,
		["camera.distanceMax"] = 32,
		["lod.textureLodMin"] = 1,
		["gfx.jpgdstfmt"] = 0,
		["lod.textureFilter"] = 2,
		["lod.landLod"] = 1,
		["lod.clutterDistance"] = 64,
		["lod.clutterDensity"] = 1,
		["draw.shadows"] = true,
		["lod.shadowMapSize"] = 2048,
		["lod.renderTargetScale"] = 1,
		["fxaa.preset"] = 3,
		["fxaa.enable"] = true,
		["world.propScreenHeightPercentMin"] = 8,
		["particle.envParticleScale"] = 0.5,
		["spell.visualSuppression"] = 0.5,
		["spell.visualSuppressedAlpha"] = 0,
		["ui.enableAddonProfiling"] = false,
	},
	["Low"] =
	{
		["video.framerateMax"] = 100,
		["lod.viewDistance"] = 256,
		["lod.farFogDistance"] = 1536,
		["camera.distanceMax"] = 32,
		["lod.textureLodMin"] = 2,
		["gfx.jpgdstfmt"] = 0,
		["lod.textureFilter"] = 1,
		["lod.landLod"] = 0,
		["lod.clutterDistance"] = 48,
		["lod.clutterDensity"] = 0,
		["draw.shadows"] = true,
		["lod.shadowMapSize"] = 1024,
		["lod.renderTargetScale"] = 0.75,
		["fxaa.preset"] = 1,
		["fxaa.enable"] = true,
		["world.propScreenHeightPercentMin"] = 12,
		["particle.envParticleScale"] = 0.25,
		["spell.visualSuppression"] = 0.9,
		["spell.visualSuppressedAlpha"] = 0,
		["ui.enableAddonProfiling"] = false,
	},
	["UltraLow"] =
	{
		["video.framerateMax"] = 100,
		["lod.viewDistance"] = 256,
		["lod.farFogDistance"] = 1024,
		["camera.distanceMax"] = 32,
		["lod.textureLodMin"] = 2,
		["gfx.jpgdstfmt"] = 0,
		["lod.textureFilter"] = 0,
		["lod.landLod"] = 0,
		["lod.clutterDistance"] = 32,
		["lod.clutterDensity"] = 0,
		["draw.shadows"] = false,
		["lod.shadowMapSize"] = 1024,
		["lod.renderTargetScale"] = 0.5,
		["fxaa.preset"] = 0,
		["fxaa.enable"] = false,
		["world.propScreenHeightPercentMin"] = 12,
		["particle.envParticleScale"] = 0.1,
		["spell.visualSuppression"] = 1.0,
		["spell.visualSuppressedAlpha"] = 0,
		["ui.enableAddonProfiling"] = false,
	}
}

local GetCurrentZoneMap = GameLib.GetCurrentZoneMap
local GetConsoleVariable = Apollo.GetConsoleVariable
local SetConsoleVariable = Apollo.SetConsoleVariable

local targetSettings = VIDEO_PRESETS_DX11.UltraLow
local targetViewDistance = 0
 
-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function QtPerformance:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 

    -- initialize variables here

    return o
end

function QtPerformance:Init()
	Apollo.RegisterAddon(self, true, "QtPerformance", {})
	self.config = {}
end
 
function QtPerformance:OnLoad()
	self.xmlDoc = XmlDoc.CreateFromFile("QtPerformance.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
	_G["qtp"] = self
  	if (GameLib.GetPlayerUnit()) then
	    self:OnCharacterCreated()
  	else
	    Apollo.RegisterEventHandler("CharacterCreated", "OnCharacterCreated", self)
  	end
end

function QtPerformance:OnDocLoaded()
	Apollo.RegisterSlashCommand("qtperformance", "OnSlashCommand", self)
	Apollo.RegisterSlashCommand("qtp", "OnSlashCommand", self)
end

function QtPerformance:OnCharacterCreated()
	if (not self.variablesLoaded) then
		self.timerGetPlayerUnit = ApolloTimer.Create(0.5, false, "OnCharacterCreated", self)
		self.variablesLoaded = true
		return
	end
	
	if (self.config == nil or self:Count(self.config) == 0) then
		self:Print("Loading default configuration.")
		self.config = self:GetDefaults()
	end
	self:ValidateConfiguration()
	if (self.config.firstRun or (self.config.savedGraphicsSettings == nil or self:Count(self.config.savedGraphicsSettings) == 0)) then
    	self:SaveCurrentConfiguration()
    	self.config.firstRun = false
    	self:OnConfigure()
    end

	self.timerApplySettings = ApolloTimer.Create(2, false, "OnApplySettings", self)
	self.timerApplySettings:Stop()
	self.timerIncreaseViewDistance = ApolloTimer.Create(0.10, true, "OnViewDistanceTimer", self)
	self.timerIncreaseViewDistance:Stop()

	Apollo.RegisterEventHandler("ChangeWorld", "OnZoneChanged", self)
    Apollo.RegisterEventHandler("SubZoneChanged", "OnZoneChanged", self)
    Apollo.RegisterEventHandler("LogOut", "OnApplySettings", self)
    self.timerApplySettings:Start()
end

function QtPerformance:OnSave(saveLevel)
  	if saveLevel ~= GameLib.CodeEnumAddonSaveLevel.Account then
	    return nil
  	end

  	return self.config
end

function QtPerformance:OnRestore(saveLevel, savedVars)
  	if saveLevel ~= GameLib.CodeEnumAddonSaveLevel.Account then
	    return
  	end

  	if savedVars ~= nil then
	    for k, v in pairs(savedVars) do
      	self.config[k] = v
    	end
  	end

  	self.variablesLoaded = true
end

function QtPerformance:ValidateConfiguration()
  	for k,v in pairs(self:GetDefaults()) do
	  	if (self.config[k] == nil) then
      	self:Print("Updating configuration (adding default value for \""..k.."\")")
      	self.config[k] = v
    	end
  	end
end

function QtPerformance:GetDefaults()
	local t = {
		firstRun = true,
		enabled = false,
		lowQualityZones = {
			"Thayd",
			"Illium",
		},
		savedGraphicsSettings = {},
		managedSettings = {
			["lod.viewDistance"] = true,
			["lod.farFogDistance"] = true,
			["lod.textureLodMin"] = false,
			["gfx.jpgdstfmt"] = false,
			["lod.textureFilter"] = false,
			["lod.landLod"] = false,
			["lod.clutterDistance"] = true,
			["lod.clutterDensity"] = true,
			["draw.shadows"] = true,
			["lod.shadowMapSize"] = true,
			["lod.renderTargetScale"] = false,
			["fxaa.preset"] = true,
			["fxaa.enable"] = true,
			["world.propScreenHeightPercentMin"] = true,
			["particle.envParticleScale"] = true,
			["spell.visualSuppression"] = true,
			["spell.visualSuppressedAlpha"] = true,
			["ui.enableAddonProfiling"] = true,
		}
	}
	return t
end

-----------------------------------------------------------------------------------------------
-- QtPerformance Functions
-----------------------------------------------------------------------------------------------
function QtPerformance:OnSlashCommand()
	self:OnConfigure()
end

function QtPerformance:OnZoneChanged()
	if (not self.config.enabled) then
		return
	end
	local currentMap = GetCurrentZoneMap()

	if (currentMap) then
		if (self:InLowGraphicsZone()) then
			if (not self.inLowGraphicsZone) then
				self.timerApplySettings:Start()
			else
			end
			self.inLowGraphicsZone = true
		else
			if (self.inLowGraphicsZone) then
				self.timerApplySettings:Start()
			else
			end
			self.inLowGraphicsZone = false
		end
	end
end

function QtPerformance:OnApplySettings()
	if (not self.config.enabled) then
		return
	end
	if (self:InLowGraphicsZone()) then
		self:ApplyConsoleVariables(targetSettings)
		self.inLowGraphicsZone = true
	else
		self:ApplyConsoleVariables(self.config.savedGraphicsSettings)
		self.inLowGraphicsZone = false
	end
end

function QtPerformance:ApplyConsoleVariables(table)
	for k,v in pairs(table) do
		--If reverting to high quality, ignore managed settings and just apply all console variables regardless.
		--If dropping to low quality, we ensure that the particular console variable is set to be managed before changing it.
		if (GetConsoleVariable(k) ~= v and (table == self.config.savedGraphicsSettings or self.config.managedSettings[k])) then
			if (k == "lod.viewDistance") then
				self.targetViewDistance = v
				self.timerIncreaseViewDistance:Start()
			else
				SetConsoleVariable(k, v)
			end
		end
	end
end

function QtPerformance:SaveCurrentConfiguration()
	self:Print("Current graphics configuration saved.  QtPerformance will revert to this configuration when you leave a low quality zone.")
	self.config.savedGraphicsSettings = {}
	for k,v in pairs(self.config.managedSettings) do
		self.config.savedGraphicsSettings[k] = GetConsoleVariable(k)
	end
end

function QtPerformance:InLowGraphicsZone()
	local zone = GetCurrentZoneMap().strName
	local found = false
	for _,v in pairs(self.config.lowQualityZones) do
		if (v == zone) then 
			found = true 
			break
		end
	end
	return found
end

function QtPerformance:OnViewDistanceTimer()
	if (GetConsoleVariable("lod.viewDistance") >= self.targetViewDistance) then
		SetConsoleVariable("lod.viewDistance", self.targetViewDistance)
		self.targetViewDistance = 0
		self.timerIncreaseViewDistance:Stop()
		return
	end
	local desiredViewDistance = tonumber(GetConsoleVariable("lod.viewDistance")) + 8
	--self:Print("Setting view distance to "..desiredViewDistance)
	SetConsoleVariable("lod.viewDistance", desiredViewDistance)
end

-----------------------------------------------------------------------------------------------
-- Config UI
-----------------------------------------------------------------------------------------------
function QtPerformance:OnConfigure()
  if (self.configWindows == nil) then
    self.configWindows = {}
    self.configWindows.config = Apollo.LoadForm(self.xmlDoc, "QtPerfPrefs", nil, self)
    self.configWindows.generalTab = self.configWindows.config:FindChild("GeneralTab")
    self.configWindows.zonesTab = self.configWindows.config:FindChild("ZonesTab")
    self.configWindows.generalBtn = self.configWindows.config:FindChild("GeneralBtn")
    self.configWindows.zonesBtn = self.configWindows.config:FindChild("ZonesBtn")

    self.checkboxMapping = {
      {wnd = self.configWindows.config:FindChild("ViewDistance"),   	savedVars = {"lod.viewDistance"}},
      {wnd = self.configWindows.config:FindChild("FogDistance"),   		savedVars = {"lod.farFogDistance"}},
      {wnd = self.configWindows.config:FindChild("ClutterDistance"),   	savedVars = {"lod.clutterDistance"}},
      {wnd = self.configWindows.config:FindChild("TextureQuality"),   	savedVars = {"lod.textureLodMin", "gfx.jpgdstfmt"}},
      {wnd = self.configWindows.config:FindChild("TextureFiltering"),   savedVars = {"lod.textureFilter"}},
      {wnd = self.configWindows.config:FindChild("Shadows"),   			savedVars = {"draw.shadows", "lod.shadowMapSize"}},
      {wnd = self.configWindows.config:FindChild("Fxaa"),   			savedVars = {"fxaa.preset", "fxaa.enable"}},
      {wnd = self.configWindows.config:FindChild("ClutterDensity"),   	savedVars = {"lod.clutterDensity"}},
      {wnd = self.configWindows.config:FindChild("SmallObjectDetail"),  savedVars = {"world.propScreenHeightPercentMin"}},
      {wnd = self.configWindows.config:FindChild("ParticleScale"),   	savedVars = {"particle.envParticleScale"}},
      {wnd = self.configWindows.config:FindChild("GroupSpellEffects"),  savedVars = {"spell.visualSuppression", "spell.visualSuppressionAlpha"}},
      {wnd = self.configWindows.config:FindChild("LandLod"),   			savedVars = {"lod.landLod"}},
      {wnd = self.configWindows.config:FindChild("RenderResolution"),   savedVars = {"lod.renderTargetScale"}},
      {wnd = self.configWindows.config:FindChild("AddonProfiling"),   	savedVars = {"ui.enableAddonProfiling"}},
    }
  end
  self.configWindows.config:Show(true)
  self:InitializeControlValues()
  self:InitializeZoneTab()
  self:SelectTab(1)
end

function QtPerformance:InitializeControlValues()
	for _,v in pairs(self.checkboxMapping) do
    	v.wnd:SetCheck(self.config.managedSettings[v.savedVars[1]])
  	end
  	self.configWindows.config:FindChild("EnableAddon"):SetCheck(self.config.enabled)
end

function QtPerformance:InitializeZoneTab()
	self.configWindows.config:FindChild("ZoneNameInput"):SetText(self.config.lowQualityZones[1])
	local grid = self.configWindows.config:FindChild("ZoneGrid")
	local currentScrollPos = grid:GetVScrollPos()
	grid:DeleteAll()
	for i=1,#self.config.lowQualityZones do
	grid:AddRow(self.config.lowQualityZones[i], "", i)
	end
	grid:SetCurrentRow(1)
	self.configWindows.config:FindChild("DeleteZoneButton"):Enable((#self.config.lowQualityZones > 1))
	self:OnZoneSelectChanged()
	grid:SetVScrollPos(currentScrollPos)
	self:OnZoneChanged()
end

function QtPerformance:OnEnabledChanged()
	self.config.enabled = self.configWindows.config:FindChild("EnableAddon"):IsChecked()
	if (not self.config.enabled) then
		self:ApplyConsoleVariables(self.config.savedGraphicsSettings)
	else
		self:OnApplySettings()
	end
end

function QtPerformance:OnManagedSettingCheckboxChanged(wndHandler, wndControl)
  	for _,v in pairs(self.checkboxMapping) do
	    if (v.wnd == wndControl) then
    		for _,v2 in pairs(v.savedVars) do
	    		self.config.managedSettings[v2] = wndControl:IsChecked()
    		end
    	end
  	end
  	if (self.config.enabled) then
	  	if (self:InLowGraphicsZone()) then
	  		for k,v in pairs(self.config.managedSettings) do
	  			if (v) then
	  				SetConsoleVariable(k, targetSettings[k])
	  			else
					SetConsoleVariable(k, self.config.savedGraphicsSettings[k])
	  			end
	  		end
	  	end
	end
end

function QtPerformance:SelectTab(index)
	if (index == 1) then
		self.configWindows.generalTab:Show(true)
		self.configWindows.generalBtn:SetCheck(true)
		self.configWindows.zonesTab:Show(false)
		self.configWindows.zonesBtn:SetCheck(false)
	else
		self.configWindows.generalTab:Show(false)
		self.configWindows.generalBtn:SetCheck(false)
		self.configWindows.zonesTab:Show(true)
		self.configWindows.zonesBtn:SetCheck(true)
	end
end

function QtPerformance:OnGeneralTabClick()
	self:SelectTab(1)
end

function QtPerformance:OnZonesTabClick()
	self:SelectTab(2)
end

function QtPerformance:OnSaveConfigButton()
	if (self.config.enabled and self:InLowGraphicsZone()) then
		self:Print("Graphics settings are currently reduced because you are in a low quality zone.  QtPerformance can't save your current configuration until you leave the area or disable the addon first.")
		return
	end
	self:SaveCurrentConfiguration()
end

function QtPerformance:OnAddZoneButton()
  table.insert(self.config.lowQualityZones, "New Zone")
  self:InitializeZoneTab()
  self.configWindows.config:FindChild("ZoneGrid"):SetCurrentRow(#self.config.lowQualityZones)
  self:OnZoneSelectChanged()
end

function QtPerformance:OnEditZoneButton()
  local row = self.configWindows.config:FindChild("ZoneGrid"):GetCurrentRow()
  local text = self.configWindows.config:FindChild("ZoneNameInput"):GetText()
  if (text == "") then
    text = self.config.lowQualityZones[row]
  end
  if (text:len() >= 200) then
    text = text:sub(1,200)
  end
  self.config.lowQualityZones[row] = text
  self:InitializeZoneTab()
  self.configWindows.config:FindChild("ZoneGrid"):SetCurrentRow(row)
  self:OnZoneSelectChanged()
end

function QtPerformance:OnDeleteZoneButton()
  local index = self.configWindows.config:FindChild("ZoneGrid"):GetCurrentRow()
  if (#self.config.lowQualityZones <= 1) then
    self:Print("This zone can't be deleted because it is the only zone in the list.")
    return
  end
  table.remove(self.config.lowQualityZones, index)
  self:InitializeZoneTab()
  if (index == 1) then index = 2 end
  self.configWindows.config:FindChild("ZoneGrid"):SetCurrentRow(index - 1)
  self:OnZoneSelectChanged()
end

function QtPerformance:OnZoneSelectChanged()
  local grid = self.configWindows.config:FindChild("ZoneGrid")
  local enableButton = #self.config.lowQualityZones > 1
  self.configWindows.config:FindChild("DeleteZoneButton"):Enable(enableButton)
  self.configWindows.config:FindChild("EditZoneButton"):Enable(true)
  self.configWindows.config:FindChild("ZoneNameInput"):SetText(self.config.lowQualityZones[grid:GetCurrentRow()])
end

function QtPerformance:OnConfigCloseBtn()
  self.configWindows.config:Close()
end

function QtPerformance:OnConfigClosed()
  self.configWindows.config:Destroy()
  self.configWindows = nil
end

-----------------------------------------------------------------------------------------------
-- Utility
-----------------------------------------------------------------------------------------------

function QtPerformance:Print(string)
  	ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_System, "[QtPerformance] "..tostring(string), "")
end

function QtPerformance:Count(table)
	if (table == nil) then
		return 0
	end
  	local count = 0
  	for _ in pairs(table) do count = count + 1 end
  	return count
end

-----------------------------------------------------------------------------------------------
-- QtPerformance Instance
-----------------------------------------------------------------------------------------------
local QtPerformanceInst = QtPerformance:new()
QtPerformanceInst:Init()
